;; vcs ver 3


;;========================== Common functions ======================================

;; -    doVOC       ( -- )
;              Run time action of VOCABULARY's.

                COLON   COMPO+5,doVOC,DOVOC
                DC      RFROM,CNTXT,STORE,EXIT

;; -    FORTH       ( -- )
;		Make FORTH the context vocabulary.

                COLON   5,FORTH,FORTH
                DC      DOVOC
                DC      0                       ;vocabulary head pointer
                DC      0                       ;vocabulary link pointer

;; -    ?DUP        ( w -- w w | 0 )    D
;		Dup tos if its is not zero.

		COLON	4,?DUP,QDUP
                DC      DUPP
                DC      QBRAN,QDUP1
                DC      DUPP
QDUP1	        DC      EXIT

;; -    ROT         ( w1 w2 w3 -- w2 w3 w1 )    D
;		Rot 3rd item to top.

		COLON	3,ROT,ROT
                DC      TOR,SWAP,RFROM,SWAP,EXIT

;; -    2DROP       ( w w -- )          D
;		Discard two items on stack.

		COLON	5,2DROP,DDROP
                DC      DROP,DROP,EXIT

;; -    2DUP        ( w1 w2 -- w1 w2 w1 w2 )    D
;		Duplicate top two items.

		COLON	4,2DUP,DDUP
                DC      OVER,OVER,EXIT

;; -    +           ( w w -- sum )      D
;		Add top two items.

		COLON	1,+,PLUS
                DC      UPLUS,DROP,EXIT

;; -    D+          ( d d -- d )
;		Double addition, as an example using UM+.
;
;		$COLON	2,'D+',DPLUS
;               DC      TOR,SWAP,TOR,UPLUS
;               DC      RFROM,RFROM,PLUS,PLUS,EXIT

;; -    NOT         ( w -- w )  D
;		One's complement of tos.

		COLON	3,NOT,INVER
                DC      DOLIT,-1,XORR,EXIT

;; -    NEGATE      ( n -- -n )         D
;		Two's complement of tos.

		COLON	6,NEGATE,NEGAT
                DC      INVER,DOLIT,1,PLUS,EXIT

;; -    DNEGATE     ( d -- -d )         D
;		Two's complement of top double.

		COLON	7,DNEGATE,DNEGA
                DC      INVER,TOR,INVER
                DC      DOLIT,1,UPLUS
                DC      RFROM,PLUS,EXIT

;; -    -           ( n1 n2 -- n1-n2 )  D
;		Subtraction.

		COLON	1,-,SUBB
                DC      NEGAT,PLUS,EXIT

;; -    ABS         ( n -- n )  D, REVIEW USING NEGATIVES EX:FAAAAA
;		Return the absolute value of n.

		COLON	3,ABS,ABSS
                DC      DUPP,ZLESS
                DC      QBRAN,ABS1
                DC      NEGAT
ABS1           DC      EXIT

;; -    =           ( w w -- t )        D
;		Return true if top two are equal.

		COLON	1,=,EQUAL
                DC      XORR
                DC      QBRAN,EQU1
                DC      DOLIT,0,EXIT            ; false flag
EQU1            DC      DOLIT,-1,EXIT           ;true flag

;; -    U<          ( u u -- t )        D
;		Unsigned compare of top two items.

		COLON	2,U<,ULESS
                DC      DDUP,XORR,ZLESS
                DC      QBRAN,ULES1
                DC      SWAP,DROP,ZLESS,EXIT
ULES1           DC      SUBB,ZLESS,EXIT

;; -    <           ( n1 n2 -- t )      D
;		Signed compare of top two items.

; :  < ( n n -- t ) 2DUP XOR 0< IF DROP 0< EXIT THEN - 0< ;

		COLON	1,<,LESS
                DC      DDUP,XORR,ZLESS
                DC      QBRAN,LESS1
                DC      DROP,ZLESS,EXIT
LESS1           DC      SUBB,ZLESS,EXIT

;; -    MAX         ( n n -- n )        D
;		Return the greater of two top stack items.

		COLON	3,MAX,MAX
                DC      DDUP,LESS
                DC      QBRAN,MAX1
                DC      SWAP
MAX1            DC      DROP,EXIT

;; -    MIN         ( n n -- n )        D
;		Return the smaller of top two stack items.

		COLON	3,MIN,MIN
                DC      DDUP,SWAP,LESS
                DC      QBRAN,MIN1
                DC      SWAP
MIN1            DC      DROP,EXIT

;; -    WITHIN      ( u ul uh -- t )    D
;		Return true if u is within the range of ul and uh.

		COLON	6,WITHIN,WITHI
                DC      OVER,SUBB,TOR                   ;ul <= u < uh
                DC      SUBB,RFROM,ULESS,EXIT



;;=================================== Divide =========================================

;; -    UM/MOD      ( udl udh u -- ur uq )
;		Unsigned divide of a double by a single. Return mod and quotient.


                COLON    6,UM/MOD,UMMOD
                DC      DDUP,ULESS
                DC      QBRAN,UMM4
                DC      NEGAT,DOLIT,23,TOR      ;!! CHANGED 15 TO 23 CELL SIZE
UMM1            DC      TOR,DUPP,UPLUS
                DC      TOR,TOR,DUPP,UPLUS
                DC      RFROM,PLUS,DUPP
                DC      RFROM,RAT,SWAP,TOR
                DC      UPLUS,RFROM,ORR
                DC      QBRAN,UMM2
                DC      TOR,DROP,DOLIT,1,PLUS,RFROM
                DC      BRAN,UMM3
UMM2            DC      DROP
UMM3            DC      RFROM
                DC      DONXT,UMM1
                DC      DROP,SWAP,EXIT
UMM4            DC      DROP,DDROP
                DC      DOLIT,-1,DUPP,EXIT      ;overflow, return max

;; -    M/MOD       ( d n -- r q )
;		Signed floored divide of double by single. Return mod and quotient.

                COLON   5,M/MOD,MSMOD
                DC      DUPP,ZLESS,DUPP,TOR
                DC      QBRAN,MMOD1
                DC      NEGAT,TOR,DNEGA,RFROM
MMOD1           DC      TOR,DUPP,ZLESS
                DC      QBRAN,MMOD2
                DC      RAT,PLUS
MMOD2           DC      RFROM,UMMOD,RFROM
                DC      QBRAN,MMOD3
                DC      SWAP,NEGAT,SWAP
MMOD3           DC      EXIT

;; -    /MOD        ( n n -- r q )
;		Signed divide. Return mod and quotient.

		COLON	4,/MOD,SLMOD
                DC      OVER,ZLESS,SWAP,MSMOD,EXIT

;; -    MOD         ( n n -- r )
;		Signed divide. Return mod only.

		COLON	3,MOD,MODD
                DC      SLMOD,DROP,EXIT

;; -    /           ( n n -- q )
;		Signed divide. Return quotient only.

		COLON	1,/,SLASH
                DC      SLMOD,SWAP,DROP,EXIT

;;================================== Multiply ========================================

;; -    UM*         ( u u -- ud )
;		Unsigned multiply. Return double product.


                COLON   3,UM*,UMSTA
                DC      DOLIT,0,SWAP,DOLIT,23,TOR       ; !! CHANGED  15 TO 23 CELL SIZE
UMST1           DC      DUPP,UPLUS,TOR,TOR
                DC      DUPP,UPLUS,RFROM,PLUS,RFROM
                DC      QBRAN,UMST2
                DC      TOR,OVER,UPLUS,RFROM,PLUS
UMST2           DC      DONXT,UMST1
                DC      ROT,DROP,EXIT

;; -    *           ( n n -- n )
;		Signed multiply. Return single product.

		COLON	1,*,STAR
                DC      UMSTA,DROP,EXIT

;; -    M*          ( n n -- d )
;		Signed multiply. Return double product.

		COLON	2,M*,MSTAR
                DC      DDUP,XORR,ZLESS,TOR
                DC      ABSS,SWAP,ABSS,UMSTA
                DC      RFROM
                DC      QBRAN,MSTA1
                DC      DNEGA
MSTA1           DC      EXIT

;; -    */MOD       ( n1 n2 n3 -- r q )
;		Multiply n1 and n2, then divide by n3. Return mod and quotient.

		COLON	5,*/MOD,SSMOD
                DC      TOR,MSTAR,RFROM,MSMOD,EXIT

;; -    */          ( n1 n2 n3 -- q )
;		Multiply n1 by n2, then divide by n3. Return quotient only.

		COLON	2,*/,STASL
                DC      SSMOD,SWAP,DROP,EXIT

;;============================ System and user variables =============================

;; -    doVAR       ( -- a )
;		Run time routine for VARIABLE and CREATE.

		COLON	COMPO+5,doVAR,DOVAR
		DC	RFROM,EXIT

;; -    doUSER      ( -- a )
;		Run time routine for user variables.

		COLON	COMPO+6,doUSER,DOUSE
		DC	RFROM,AT,UP,AT,PLUS,EXIT

;; -    UP          ( -- a )
;		Pointer to the user area.

		COLON	2,UP,UP
		DC	DOVAR
		DC	UPP


;; -    SP0         ( -- a )
;		Pointer to bottom of the data stack.

		USER	3,SP0,SZERO

;; -    RP0         ( -- a )
;		Pointer to bottom of the return stack.

		USER	3,RP0,RZERO


;; -    '?KEY       ( -- a )
;		Execution vector of ?KEY.

                HCODE    5,TQKEY
                DC      $27
		DC	'?KEY'
                HUSER

;; -    'EMIT       ( -- a )
;		Execution vector of EMIT.

                HCODE   5,TEMIT
                DC      $27
		DC	'EMIT'
                HUSER

;; -    'EXPECT     ( -- a )
;		Execution vector of EXPECT.
                HCODE   7,TEXPE
                DC      $27
		DC	'EXPECT'
                HUSER

;; -    'TAP        ( -- a )
;		Execution vector of TAP.
                HCODE   4,TTAP
                DC      $27
		DC	'TAP'
                HUSER

;; -    'ECHO       ( -- a )
;		Execution vector of ECHO.
                HCODE   5,TECHO
                DC      $27
		DC	'ECHO'
                HUSER

;; -    'PROMPT     ( -- a )
;		Execution vector of PROMPT.

                HCODE   7,TPROM
                DC      $27
		DC	'PROMPT'
                HUSER

;; -    BASE        ( -- a )
;		Storage of the radix base for numeric I/O.

		USER	4,BASE,BASE

;; -    MEMSPACE       ( -- a )
                USER    8,MEMSPACE,MEMSPAC


;; -    tmp         ( -- a )
;		A temporary storage location used in parse and find.

		USER	COMPO+3,tmp,TEMP

;; -    SPAN        ( -- a )
;		Hold character count received by EXPECT.

		USER	4,SPAN,SPAN

;; -    >IN         ( -- a )
;		Hold the character pointer while parsing input stream.

		USER	3,>IN,INN

;; -    #TIB        ( -- a )
;		Hold the current count and address of the terminal input buffer.

		USER	4,#TIB,NTIB
USERp set USERp+CELLL		;hold the base address of the terminal input buffer

;; -    CSP         ( -- a )
;		Hold the stack pointer for error checking.

		USER	3,CSP,CSP

;; -    'EVAL       ( -- a )
;		Execution vector of EVAL.


                HCODE   5,TEVAL
                DC      $27
		DC	'EVAL'
                HUSER


;; -    'NUMBER     ( -- a )
;		Execution vector of NUMBER?.
                HCODE    7,TNUMB
                DC      $27
		DC	'NUMBER'
                HUSER

;; -    HLD         ( -- a )
;		Hold a pointer in building a numeric output string.

		USER	3,HLD,HLD

;; -    HANDLER     ( -- a )
;		Hold the return stack pointer for error handling.

                USER    7,HANDLER,HANDL

;; -    CONTEXT     ( -- a )
;		A area to specify vocabulary search order.

		USER	7,CONTEXT,CNTXT
USERp = USERp+(VOCSS*CELLL)       ;vocabulary stack

;; -    CURRENT     ( -- a )
;		Point to the vocabulary to be extended.

		USER	7,CURRENT,CRRNT
USERp = USERp+CELLL		;vocabulary link pointer

;; -    CP          ( -- a )
;		Point to the top of the code dictionary.

		USER	2,CP,CP

;; -    NP          ( -- a )
;		Point to the bottom of the name dictionary.

		USER	2,NP,NP




;; -    LAST        ( -- a )
;		Point to the last name in the name dictionary.

		USER	4,LAST,LAST



;;================================= Miscellaneous =====================================

;; -    CELL+       ( a -- a )
;		Add cell size in byte to address.

		COLON	5,CELL+,CELLP
                DC      DOLIT,CELLL,PLUS,EXIT

;; -    CELL-       ( a -- a )
;		Subtract cell size in byte from address.

		COLON	5,CELL-,CELLM
                DC      DOLIT,0-CELLL,PLUS,EXIT

;; -    CELLS       ( n -- n )
;		Multiply tos by cell size in bytes.

		COLON	5,CELLS,CELLS
                DC      DOLIT,CELLL,STAR,EXIT

;; -    ALIGNED     ( b -- a )
;		Align address to the cell boundary.

		COLON	7,ALIGNED,ALGND
                DC      DUPP,DOLIT,0,DOLIT,CELLL
                DC      UMMOD,DROP,DUPP
                DC      QBRAN,ALGN1
                DC      DOLIT,CELLL,SWAP,SUBB
ALGN1:          DC      PLUS,EXIT

;; -    BL          ( -- 32 )
;		Return 32, the blank character.

		COLON	2,BL,BLANK
                DC      DOLIT,' ',EXIT

;; -    >CHAR       ( c -- c )
;		Filter non-printing characters.

		COLON	5,>CHAR,TCHAR
                DC      DOLIT,$07F,ANDD,DUPP    ;mask msb
                DC      DOLIT,127,BLANK,WITHI   ;check for printable
                DC      QBRAN,TCHA1
                DC      DROP,DOLIT,'_'          ;replace non-printables
TCHA1           DC      EXIT

;; -    DEPTH       ( -- n )
;		Return the depth of the data stack.

		COLON	5,DEPTH,DEPTH
                DC      SPAT,SZERO,AT,SWAP,SUBB
                DC      DOLIT,CELLL,SLASH,EXIT

;; -    PICK        ( ... +n -- ... w )
;		Copy the nth stack item to tos.

		COLON	4,PICK,PICK
                DC      DOLIT,1,PLUS,CELLS
                DC      SPAT,PLUS,AT,EXIT

;;================================= Memory access ===================================

;; -    +!          ( n a -- )
;		Add n to the contents at address a.

		COLON	2,+!,PSTOR
                DC      SWAP,OVER,AT,PLUS
                DC      SWAP,STORE,EXIT

;; -    2!          ( d a -- )
;		Store the double integer to address a.

		COLON	2,2!,DSTOR
                DC      SWAP,OVER,STORE
                DC      CELLP,STORE,EXIT

;; -    2@          ( a -- d )
;		Fetch double integer from address a.

		COLON	2,2@,DAT
                DC      DUPP,CELLP,AT
                DC      SWAP,AT,EXIT

;; -    COUNT       ( b -- b +n )
;		Return count byte of a string and add 1 to byte address.

		COLON	5,COUNT,COUNT
                DC      DUPP,DOLIT,1,PLUS
                DC      SWAP,CAT,EXIT

;; -    HERE        ( -- a )
;		Return the top of the code dictionary.

		COLON	4,HERE,HERE
                DC      CP,AT,EXIT

;; -    PAD         ( -- a )
;		Return the address of a temporary buffer.

		COLON	3,PAD,PAD
                DC      HERE,DOLIT,80,PLUS,EXIT

;; -    TIB         ( -- a )
;		Return the address of the terminal input buffer.

		COLON	3,TIB,TIB
                DC      NTIB,CELLP,AT,EXIT

;; -    @EXECUTE    ( a -- )
;		Execute vector stored in address a.

		COLON	8,@EXECUTE,ATEXE
                DC      AT,QDUP                 ;?address or zero
                DC      QBRAN,EXE1
                DC      EXECU                   ;execute if non-zero
EXE1:           DC      EXIT                    ;do nothing if zero

;; -    CMOVE       ( b1 b2 u -- )
;		Copy u bytes from b1 to b2.

		COLON	5,CMOVE,CMOVE
                DC      TOR
                DC      BRAN,CMOV2
CMOV1          DC      TOR,DUPP,CAT
                DC      RAT,CSTOR
                DC      DOLIT,1,PLUS
                DC      RFROM,DOLIT,1,PLUS
CMOV2          DC      DONXT,CMOV1
                DC      DDROP,EXIT

;; -    FILL        ( b u c -- )
;		Fill u bytes of character c to area beginning at b.

		COLON	4,FILL,FILL
                DC      SWAP,TOR,SWAP
                DC      BRAN,FILL2
FILL1           DC      DDUP,CSTOR,DOLIT,1,PLUS
FILL2           DC      DONXT,FILL1
                DC      DDROP,EXIT

;; -    -TRAILING   ( b u -- b u )
;		Adjust the count to eliminate trailing white space.

		COLON	9,-TRAILING,DTRAI
                DC      TOR
                DC      BRAN,DTRA2
DTRA1           DC      BLANK,OVER,RAT,PLUS,CAT,LESS
                DC      QBRAN,DTRA2
                DC      RFROM,DOLIT,1,PLUS,EXIT ;adjusted count
DTRA2           DC      DONXT,DTRA1
                DC      DOLIT,0,EXIT            ;count=0

;; -    PACK$       ( b u a -- a )
;		Build a counted string with u characters from b. Null fill.

		COLON	5,PACK$,PACKS
PACK1           DC      ALGND,DUPP,TOR          ;strings only on cell boundary
                DC      OVER,DUPP,DOLIT,0
                DC      DOLIT,CELLL,UMMOD,DROP  ;count mod cell
                DC      SUBB,OVER,PLUS
                DC      DOLIT,0,SWAP,STORE      ;null fill cell
                DC      DDUP,CSTOR,DOLIT,1,PLUS ;save count
                DC      SWAP,CMOVE,RFROM,EXIT   ;move string

;;========================= Numeric output, single precision ========================
;.( Numeric Output ) \ single precision
;: HOLD ( c -- ) HLD @ 1 - DUP HLD ! C! ;
;: SIGN ( n -- ) 0< IF [ CHAR - ] LITERAL HOLD THEN ;
; : #> ( w -- b u ) DROP HLD @ PAD OVER - ;
;: str ( w -- b u ) DUP >R ABS <# #S R> SIGN #> ;
;: HEX ( -- ) 16 BASE ! ;
;: DECIMAL ( -- ) 10 BASE ! ;

;; -    DIGIT       ( u -- c )
;               Convert digit u to a character.
; : DIGIT ( u -- c ) 9 OVER < 7 AND + [ CHAR 0 ] LITERAL + ;

		COLON	5,DIGIT,DIGIT
                DC      DOLIT,9,OVER,LESS
                DC      DOLIT,7,ANDD,PLUS
                DC      DOLIT,'0',PLUS,EXIT

;; -    EXTRACT     ( n base -- n c )
;		Extract the least significant digit from n.
; : EXTRACT ( n base -- n c ) 0 SWAP UM/MOD SWAP DIGIT ;

		COLON	7,EXTRACT,EXTRC
                DC      DOLIT,0,SWAP,UMMOD
                DC      SWAP,DIGIT,EXIT

;; -    <#          ( -- )
;		Initiate the numeric output process.
;: <# ( -- ) PAD HLD ! ;

		COLON	2,<#,BDIGS
                DC      PAD,HLD,STORE,EXIT

;; -    HOLD        ( c -- )
;		Insert a character into the numeric output string.

; ! hld = 8C2

		COLON	4,HOLD,HOLD
                DC      HLD,AT,DOLIT,1,SUBB
                DC      DUPP,HLD,STORE,CSTOR,EXIT

;; -    #           ( u -- u )
; : # ( u -- u ) BASE @ EXTRACT HOLD ;
;		Extract one digit from u and append the digit to output string.

		COLON	1,#,DIG
                DC      BASE,AT,EXTRC,HOLD,EXIT

;; -    #S          ( u -- 0 )
;		Convert u until all digits are added to the output string.
; : #S ( u -- 0 ) BEGIN # DUP WHILE REPEAT ;

		COLON	2,#S,DIGS
DIGS1           DC      DIG,DUPP
                DC      QBRAN,DIGS2
                DC      BRAN,DIGS1
DIGS2           DC      EXIT

;; -    SIGN        ( n -- )
;		Add a minus sign to the numeric output string.

		COLON	4,SIGN,SIGN
                DC      ZLESS
                DC      QBRAN,SIGN1
                DC      DOLIT,'-',HOLD
SIGN1           DC      EXIT

;; -    #>          ( w -- b u )
;		Prepare the output string to be TYPE'd.

		COLON	2,#>,EDIGS
                DC      DROP,HLD,AT
                DC      PAD,OVER,SUBB,EXIT

;; -    str         ( n -- b u )
;		Convert a signed integer to a numeric string.

		COLON	3,str,STR
                DC      DUPP,TOR,ABSS
                DC      BDIGS,DIGS,RFROM
                DC      SIGN,EDIGS,EXIT

;; -    HEX         ( -- )
;		Use radix 16 as base for numeric conversions.

		COLON	3,HEX,HEX
                DC      DOLIT,16,BASE,STORE,EXIT

;; -    DECIMAL     ( -- )
;		Use radix 10 as base for numeric conversions.

		COLON	7,DECIMAL,DECIM
                DC      DOLIT,10,BASE,STORE,EXIT

;;============================ Numeric input, single precision ======================

;; -    DIGIT?      ( c base -- u t )
;		Convert a character to its numeric value. A flag indicates success.

		COLON	6,DIGIT?,DIGTQ
                DC      TOR,DOLIT,'0',SUBB
                DC      DOLIT,9,OVER,LESS
                DC      QBRAN,DGTQ1
                DC      DOLIT,7,SUBB
                DC      DUPP,DOLIT,10,LESS,ORR
DGTQ1           DC      DUPP,RFROM,ULESS,EXIT

;; -    NUMBER?     ( a -- n T | a F )
;		Convert a number string to integer. Push a flag on tos.

                COLON   7,NUMBER?,NUMBQ
                DC      BASE,AT,TOR,DOLIT,0,OVER,COUNT
                DC      OVER,CAT,DOLIT,'$',EQUAL
                DC      QBRAN,NUMQ1
                DC      HEX,SWAP,DOLIT,1,PLUS
                DC      SWAP,DOLIT,1,SUBB
NUMQ1:          DC      OVER,CAT,DOLIT,'-',EQUAL,TOR
                DC      SWAP,RAT,SUBB,SWAP,RAT,PLUS,QDUP
                DC      QBRAN,NUMQ6
                DC      DOLIT,1,SUBB,TOR
NUMQ2:          DC      DUPP,TOR,CAT,BASE,AT,DIGTQ
                DC      QBRAN,NUMQ4
                DC      SWAP,BASE,AT,STAR,PLUS,RFROM
                DC      DOLIT,1,PLUS
                DC      DONXT,NUMQ2
                DC      RAT,SWAP,DROP
                DC      QBRAN,NUMQ3
                DC      NEGAT
NUMQ3:          DC      SWAP
                DC      BRAN,NUMQ5
NUMQ4:          DC      RFROM,RFROM,DDROP,DDROP,DOLIT,0
NUMQ5:          DC      DUPP
NUMQ6:          DC      RFROM,DDROP
                DC      RFROM,BASE,STORE,EXIT


;;=================================== Basic I/O ====================================

;; -    ?KEY        ( -- c T | F )
;		Return input character and true, or a false if no input.

		COLON	4,?KEY,QKEY
                DC      TQKEY,ATEXE,EXIT

;; -    KEY         ( -- c )
;		Wait for and return an input character.

		COLON	3,KEY,KEY
KEY1            DC      QKEY
                DC      QBRAN,KEY1
                DC      EXIT

;; -    EMIT        ( c -- )
;		Send a character to the output device.

		COLON	4,EMIT,EMIT
                DC      TEMIT,ATEXE,EXIT

;; -    NUF?        ( -- t )
;		Return false if no input, else pause and if CR return true.

		COLON	4,NUF?,NUFQ
                DC      QKEY,DUPP
                DC      QBRAN,NUFQ1
                DC      DDROP,KEY,DOLIT,CRR,EQUAL
NUFQ1:          DC      EXIT

;; -    PACE        ( -- )
;		Send a pace character for the file downloading process.

		COLON	4,PACE,PACE
                DC      DOLIT,11,EMIT,EXIT

;; -    SPACE       ( -- )
;		Send the blank character to the output device.

		COLON	5,SPACE,SPACE
                DC      BLANK,EMIT,EXIT

;; -    SPACES      ( +n -- )
;		Send n spaces to the output device.

		COLON	6,SPACES,SPACS
                DC      DOLIT,0,MAX,TOR
                DC      BRAN,CHAR2
CHAR1           DC      SPACE
CHAR2           DC      DONXT,CHAR1
                DC      EXIT

;; -    TYPE        ( b u -- )
;		Output u characters from b.

		COLON	4,TYPE,TYPEE
                DC      TOR
                DC      BRAN,TYPE2
TYPE1           DC      DUPP,CAT,EMIT
                DC      DOLIT,1,PLUS
TYPE2           DC      DONXT,TYPE1
                DC      DROP,EXIT

;; -    CR          ( -- )
;		Output a carriage return and a line feed.

		COLON	2,CR,CR
                DC      DOLIT,CRR,EMIT
                DC      DOLIT,LF,EMIT,EXIT


;; -    do$         ( -- a )
;		Return the address of a compiled string.

		COLON	COMPO+3,do$,DOSTR
DOST1           DC      RFROM,RAT,RFROM,COUNT,PLUS
                DC      ALGND,TOR,SWAP,TOR
DOSTEX          DC      EXIT

;                ORG     X:BPPTR        ; LATER DUDE
;                DC      DOSTX

;                DC      TOR,SWAP,TOR,EXIT

;; -    $"|         ( -- a )
;		Run time routine compiled by $". Return address of a compiled string.

		HCODE	COMPO+3,STRQP
		DC	'$','"','|'
		HCOLON
		DC	DOSTR,EXIT		;force a call to do$


;; -    ."|         ( -- )
;		Run time routine of ." . Output a compiled string.

		HCODE	COMPO+3,DOTQP
		DC	'.','"','|',
		HCOLON
		DC	DOSTR,COUNT,TYPEE,EXIT




;; -    .R          ( n +n -- )
;		Display an integer in a field of n columns, right justified.

		COLON	2,.R,DOTR
                DC      TOR,STR,RFROM,OVER,SUBB
                DC      SPACS,TYPEE,EXIT

;; -    U.R         ( u +n -- )
;		Display an unsigned integer in n column, right justified.

; : U.R ( u +n -- ) >R <# #S #> R> OVER - SPACES TYPE ;

		COLON	3,U.R,UDOTR
                DC      TOR,BDIGS,DIGS,EDIGS
                DC      RFROM,OVER,SUBB
                DC      SPACS,TYPEE,EXIT

;; -    U.          ( u -- )
;		Display an unsigned integer in free format.

		COLON	2,U.,UDOT
                DC      BDIGS,DIGS,EDIGS
                DC      SPACE,TYPEE,EXIT

;; -    .           ( w -- )
;		Display an integer in free format, preceeded by a space.

		COLON	1,.,DOT
                DC      BASE,AT,DOLIT,10,XORR   ;?decimal
                DC      QBRAN,DOT1
                DC      UDOT,EXIT               ;no, display unsigned
DOT1:           DC      STR,SPACE,TYPEE,EXIT    ;yes, display signed

;; -    ?           ( a -- )
;		Display the contents in a memory cell.

		COLON	1,?,QUEST
                DC      AT,DOT,EXIT

;;================================== Parsing ======================================

;; -    parse       ( b u c -- b u delta ; <string> )
;		Scan string delimited by c. Return found string and its offset.

		COLON	5,parse,PARS
                DC      TEMP,STORE,OVER,TOR,DUPP
                DC      QBRAN,PARS8
                DC      DOLIT,1,SUBB,TEMP,AT,BLANK,EQUAL
                DC      QBRAN,PARS3
                DC      TOR
PARS1           DC      BLANK,OVER,CAT          ;skip leading blanks ONLY
                DC      SUBB,ZLESS,INVER
                DC      QBRAN,PARS2
                DC      DOLIT,1,PLUS
                DC      DONXT,PARS1
                DC      RFROM,DROP,DOLIT,0,DUPP,EXIT
PARS2           DC      RFROM
PARS3           DC      OVER,SWAP
                DC      TOR
PARS4           DC      TEMP,AT,OVER,CAT,SUBB   ;scan for delimiter
                DC      TEMP,AT,BLANK,EQUAL
                DC      QBRAN,PARS5
                DC      ZLESS
PARS5           DC      QBRAN,PARS6
                DC      DOLIT,1,PLUS
                DC      DONXT,PARS4
                DC      DUPP,TOR
                DC      BRAN,PARS7
PARS6           DC      RFROM,DROP,DUPP
                DC      DOLIT,1,PLUS,TOR
PARS7           DC      OVER,SUBB
                DC      RFROM,RFROM,SUBB,EXIT
PARS8           DC      OVER,RFROM,SUBB,EXIT

;; -    PARSE       ( c -- b u ; <string> )
;		Scan input stream and return counted string delimited by c.

		COLON	5,PARSE,PARSE
                DC      TOR
        	DC	TIB,INN,AT,PLUS     ;current input buffer pointer
                DC      NTIB,AT,INN,AT,SUBB     ;remaining count
                DC      RFROM,PARS,INN,PSTOR,EXIT

;; -    .(          ( -- )
;		Output following string up to next ) .

		COLON	IMEDD+2,.(,DOTPR
                DC      DOLIT,')',PARSE,TYPEE,EXIT

;; -    (           ( -- )
;		Ignore following string up to next ) . A comment.

		COLON	IMEDD+1,(,PAREN
                DC      DOLIT,')',PARSE,DDROP,EXIT

;; -    \           ( -- )
;		Ignore following text till the end of line.

		COLON	IMEDD+1,\,BKSLA
                DC      NTIB,AT,INN,STORE,EXIT

;; -    CHAR        ( -- c )
;		Parse next word and return its first character.

		COLON	4,CHAR,CHAR
                DC      BLANK,PARSE,DROP,CAT,EXIT

;; -    TOKEN       ( -- a ; <string> )
;		Parse a word from input stream and copy it to name dictionary.

		COLON	5,TOKEN,TOKEN
                DC      BLANK,PARSE,DOLIT,31,MIN
                DC      NP,AT,OVER,SUBB,CELLM
                DC      PACKS,EXIT

;; -    WORD        ( c -- a ; <string> )
;		Parse a word from input stream and copy it to code dictionary.

		COLON	4,WORD,WORDD
                DC      PARSE,HERE,PACKS,EXIT

;;=============================== Dictionary search ==================================

;; -    NAME>       ( na -- ca )
;		Return a code address given a name address.

		COLON	5,NAME>,NAMET
                DC      CELLM,CELLM,AT,EXIT

;; -    SAME?       ( a a u -- a a f \ -0+ )
;		Compare u cells in two strings. Return 0 if identical.

		COLON	5,SAME?,SAMEQ
                DC      TOR
                DC      BRAN,SAME2
SAME1           DC      OVER,RAT,CELLS,PLUS,AT
                DC      OVER,RAT,CELLS,PLUS,AT
                DC      SUBB,QDUP
                DC      QBRAN,SAME2
                DC      RFROM,DROP,EXIT         ;strings not equal
SAME2           DC      DONXT,SAME1
                DC      DOLIT,0,EXIT            ;strings equal

;; -    find        ( a va -- ca na | a F )
;		Search a vocabulary for a string. Return ca and na if succeeded.

		COLON	4,find,FIND
                DC      SWAP,DUPP,CAT
                DC      DOLIT,CELLL,SLASH,TEMP,STORE
                DC      DUPP,AT,TOR,CELLP,SWAP
FIND1           DC      AT,DUPP
                DC      QBRAN,FIND6
                DC      DUPP,AT,DOLIT,MASKK,ANDD,RAT,XORR
                DC      QBRAN,FIND2
                DC      CELLP,DOLIT,-1          ;true flag
                DC      BRAN,FIND3
FIND2           DC      CELLP,TEMP,AT,SAMEQ
FIND3           DC      BRAN,FIND4
FIND6           DC      RFROM,DROP
                DC      SWAP,CELLM,SWAP,EXIT
FIND4           DC      QBRAN,FIND5
                DC      CELLM,CELLM
                DC      BRAN,FIND1
FIND5           DC      RFROM,DROP,SWAP,DROP
                DC      CELLM
                DC      DUPP,NAMET,SWAP,EXIT

;; -    NAME?       ( a -- ca na | a F )
;		Search all context vocabularies for a string.

		COLON	5,NAME?,NAMEQ
                DC      CNTXT,DUPP,DAT,XORR     ;?context=also
                DC      QBRAN,NAMQ1
                DC      CELLM                   ;no, start with context
NAMQ1           DC      TOR
NAMQ2           DC      RFROM,CELLP,DUPP,TOR    ;next in search order
                DC      AT,QDUP
                DC      QBRAN,NAMQ3
                DC      FIND,QDUP               ;search vocabulary
                DC      QBRAN,NAMQ2
                DC      RFROM,DROP,EXIT         ;found name
NAMQ3           DC      RFROM,DROP              ;name not found
                DC      DOLIT,0,EXIT            ;false flag

;;============================== Terminal response ==================================

;; -    ^H          ( bot eot cur -- bot eot cur )
;		Backup the cursor by one character.

		COLON	2,^H,BKSP
                DC      TOR,OVER,RFROM,SWAP,OVER,XORR
                DC      QBRAN,BACK1
                DC      DOLIT,BKSPP,TECHO,ATEXE,DOLIT,1,SUBB
                DC      BLANK,TECHO,ATEXE
                DC      DOLIT,BKSPP,TECHO,ATEXE
BACK1:          DC      EXIT

;; -    TAP         ( bot eot cur c -- bot eot cur )
;		Accept and echo the key stroke and bump the cursor.

		COLON	3,TAP,TAP
                DC      DUPP,TECHO,ATEXE
                DC      OVER,CSTOR,DOLIT,1,PLUS,EXIT

;; -    kTAP        ( bot eot cur c -- bot eot cur )
;		Process a key stroke, CR or backspace.

		COLON	4,kTAP,KTAP
                DC      DUPP,DOLIT,CRR,XORR
                DC      QBRAN,KTAP2
                DC      DOLIT,BKSPP,XORR
                DC      QBRAN,KTAP1
                DC      BLANK,TAP,EXIT
KTAP1           DC      BKSP,EXIT
KTAP2           DC      DROP,SWAP,DROP,DUPP,EXIT

;; -    accept      ( b u -- b u )
;		Accept characters to input buffer. Return with actual count.

		COLON	6,accept,ACCEP
                DC      OVER,PLUS,OVER
ACCP1           DC      DDUP,XORR
                DC      QBRAN,ACCP4
                DC      KEY,DUPP
;               DC      BLANK,SUBB,DOLIT,95,ULESS
                DC      BLANK,DOLIT,127,WITHI
                DC      QBRAN,ACCP2
                DC      TAP
                DC      BRAN,ACCP3
ACCP2           DC      TTAP,ATEXE
ACCP3           DC      BRAN,ACCP1
ACCP4           DC      DROP,OVER,SUBB,EXIT

;; -    EXPECT      ( b u -- )
;		Accept input stream and store count in SPAN.

		COLON	6,EXPECT,EXPEC
                DC      TEXPE,ATEXE,SPAN,STORE,DROP,EXIT

;; -    QUERY       ( -- )
;		Accept input stream to terminal input buffer.

		COLON	5,QUERY,QUERY
                DC      TIB,DOLIT,80,TEXPE,ATEXE,NTIB,STORE
                DC      DROP,DOLIT,0,INN,STORE,EXIT

;;================================ Error handling ================================

;; -    CATCH       ( ca -- 0 | err# )
;		Execute word at ca and set up an error frame for it.

		COLON	5,CATCH,CATCH
                DC      SPAT,TOR,HANDL,AT,TOR   ;save error frame
                DC      RPAT,HANDL,STORE,EXECU    ;execute
                DC      RFROM,HANDL,STORE       ;restore error frame
                DC      RFROM,DROP,DOLIT,0,EXIT	;no error

;; -    THROW       ( err# -- err# )
;		Reset system to current local error frame an update error flag.

		COLON	5,THROW,THROW
                DC      HANDL,AT,RPSTO          ;restore return stack
                DC      RFROM,HANDL,STORE       ;restore handler frame
                DC      RFROM,SWAP,TOR,SPSTO    ;restore data stack
                DC      DROP,RFROM,EXIT

;; -    NULL$       ( -- a )
;		Return address of a null string with zero count.

      		HCODE	5,NULLS
		DC	'NULLL'
		DC	'$'
		HCOLON
		DC	DOVAR			;emulate CREATE
		DC	0
		DC	99
		DC	111
		DC	121
		DC	111
		DC	116
		DC	101


;; -    ABORT       ( -- )
;		Reset data stack and jump to QUIT.

		COLON	5,ABORT,ABORT
                DC      NULLS,THROW

;; -    abort"      ( f -- )
;		Run time routine of ABORT" . Abort with a message.

		HCODE	COMPO+6,ABORQ
		DC	'abort"'
		HCOLON
		DC	QBRAN,ABOR1		;text flag
		DC	DOSTR,THROW		;pass error string
ABOR1		DC	DOSTR,DROP,EXIT		;drop error


;;================================  The text interpreter ================================




;; -    $INTERPRET  ( a -- )
;		Interpret a word. If failed, try to convert it to an integer.

                COLON   10,$INTERPRET,INTER
                DC      NAMEQ,QDUP              ;?defined
                DC      QBRAN,INTE1
                DC      AT,DOLIT,COMPO,ANDD     ;?compile only lexicon bits
                DC      ABORQ
		DC	15
		DC	' compile only  '
                DC      EXECU,EXIT              ;execute defined word
INTE1           DC      TNUMB,ATEXE             ;convert a number
		DC	QBRAN,INTE2
		DC	EXIT
INTE2		DC	THROW			;error

;; -    [           ( -- )
;		Start the text interpreter.

		HCODE	IMEDD+1,LBRAC
		DC	'['
		HCOLON
                DC      DOLIT
LBINT   	DC	INTER,TEVAL,STORE,EXIT


;; -    .OK         ( -- )
;		Display 'ok' only while interpreting.

		COLON	3,.OK,DOTOK
                DC      DOLIT,INTER,TEVAL,AT,EQUAL
                DC      QBRAN,DOTO1
		DC	DOTQP
		DC	3
		DC	' ok'
DOTO1           DC      CR,EXIT

;; -    ?STACK      ( -- )
;		Abort if the data stack underflows.

		COLON	6,?STACK,QSTAC
                DC      DEPTH,ZLESS             ;check only for underflow
		DC	ABORQ,10
		DC	' underflow'
                DC      EXIT

;; -    EVAL        ( -- )
;		Interpret the input stream.

		COLON	4,EVAL,EVAL
EVAL1           DC      TOKEN,DUPP,CAT          ;?input stream empty
                DC      QBRAN,EVAL2
                DC      TEVAL,ATEXE,QSTAC       ;evaluate input, check stack
                DC      BRAN,EVAL1
EVAL2           DC      DROP,TPROM,ATEXE,EXIT   ;prompt

;;================================  Shell ================================

;; -    PRESET      ( -- )
;		Reset data stack pointer and the terminal input buffer.

		COLON	6,PRESET,PRESE
                DC      SZERO,AT,SPSTO
                DC      DOLIT,TIBB,NTIB,CELLP,STORE,EXIT

;; -    xio         ( a a a -- )
;		Reset the I/O vectors 'EXPECT, 'TAP, 'ECHO and 'PROMPT.

		COLON	COMPO+3,xio,XIO
                DC      DOLIT,ACCEP,TEXPE,DSTOR
                DC      TECHO,DSTOR,EXIT

;; -    FILE        ( -- )
;		Select I/O vectors for file download.

		COLON	4,FILE,FILE
                DC      DOLIT,PACE,DOLIT,DROP
                DC      DOLIT,KTAP,XIO,EXIT

;; -    HAND        ( -- )
;		Select I/O vectors for terminal interface.

		COLON	4,HAND,HAND
                DC      DOLIT,DOTOK,DOLIT,EMIT
                DC      DOLIT,KTAP,XIO,EXIT

;; -    I/O         ( -- a )
;		Array to store default I/O vectors.

		COLON	3,I/O,ISLO
                DC      DOVAR                   ;emulate CREATE
                DC      QRX,TXSTO               ;default I/O vectors

;; -    CONSOLE     ( -- )
;		Initiate terminal interface.

		COLON	7,CONSOLE,CONSO
                DC      ISLO,DAT,TQKEY,DSTOR    ;restore default I/O device
                DC      HAND,EXIT               ;keyboard input

;; -    QUIT        ( -- )
;		Reset return stack pointer and start text interpreter.

		COLON	4,QUIT,QUIT
                DC      RZERO,AT,RPSTO          ;reset return stack pointer
QUIT1           DC      LBRAC                   ;start interpretation
QUIT2           DC      QUERY                   ;get input
                DC      DOLIT                   ;evaluate input
                DC      EVAL
                DC      CATCH
                DC      QDUP
                DC      QBRAN,QUIT2             ;continue till error
                DC      TPROM,AT,SWAP           ;save input device
                DC      CONSO,NULLS,OVER,XORR   ;?display error message
                DC      QBRAN,QUIT3
                DC      SPACE,COUNT,TYPEE       ;error message
		DC	DOTQP
		DC	3
		DC	' ? '
QUIT3           DC      DOLIT,DOTOK,XORR        ;?file input
                DC      QBRAN,QUIT4
                DC      DOLIT,ERR,EMIT          ;file error, tell host
QUIT4           DC      PRESE                   ;some cleanup
                DC      BRAN,QUIT1

;;================================  The compiler ================================

;; -    '           ( -- ca )
;		Search context vocabularies for the next word in input stream.

;   '           ( -- ca )

                HCODE   1,TICK
                DC      $27
                HCOLON
                DC      TOKEN,NAMEQ             ;?defined
                DC      QBRAN,TICK1
                DC      EXIT                    ;yes, push code address
TICK1:          DC      THROW                   ;no, error

;; -    ALLOT       ( n -- )
;		Allocate n bytes to the code dictionary.

		COLON	5,ALLOT,ALLOT
                DC      CP,PSTOR,EXIT           ;adjust code pointer

;; -    ,           ( w -- )
;		Compile an integer into the code dictionary.

                HCODE   1,COMMA
                DC      $2C
                HCOLON
                DC      HERE,DUPP,CELLP         ;cell boundary
                DC      CP,STORE,STORE,EXIT     ;adjust code pointer, compile



;; -    [COMPILE]   ( -- ; <string> )
;		Compile the next immediate word into code dictionary.


                HCODE   IMEDD+9,BCOMP
                DC      '[COMPILE]'
                HCOLON
		DC	TICK,COMMA,EXIT



;; -    COMPILE     ( -- )
;		Compile the next address in colon list to code dictionary.

		COLON	COMPO+7,COMPILE,COMPI
                DC      RFROM,DUPP,AT,COMMA     ;compile address
                DC      CELLP,TOR,EXIT          ;adjust return address

;; -    LITERAL     ( w -- )
;		Compile tos to code dictionary as an integer literal.

		COLON	IMEDD+7,LITERAL,LITER
                DC      COMPI,DOLIT,COMMA,EXIT

;; -    $,"         ( -- )
;		Compile a literal string up to next " .

                HCODE   3,STRCQ
                DC      '$',$2C,'"'
                HCOLON
		DC	DOLIT,'"',WORDD		;move string to code dictionary
                DC      COUNT,PLUS,ALGND        ;calculate aligned end of string
		DC	CP,STORE,EXIT		;adjust the code pointer


;; -    RECURSE     ( -- )
;		Make the current word available for compilation.

		COLON	IMEDD+7,RECURSE,RECUR
                DC      LAST,AT,NAMET,COMMA,EXIT

;;================================  Structures ================================

;; -    FOR         ( -- a )
;		Start a FOR-NEXT loop structure in a colon definition.

		COLON	IMEDD+3,FOR,FOR
                DC      COMPI,TOR,HERE,EXIT

;; -    BEGIN       ( -- a )
;		Start an infinite or indefinite loop structure.

		COLON	IMEDD+5,BEGIN,BEGIN
                DC      HERE,EXIT

;; -    NEXT        ( a -- )
;		Terminate a FOR-NEXT loop structure.

		COLON	IMEDD+4,NEXT,NEXT
                DC      COMPI,DONXT,COMMA,EXIT

;; -    UNTIL       ( a -- )
;		Terminate a BEGIN-UNTIL indefinite loop structure.

		COLON	IMEDD+5,UNTIL,UNTIL
                DC      COMPI,QBRAN,COMMA,EXIT

;; -    AGAIN       ( a -- )
;		Terminate a BEGIN-AGAIN infinite loop structure.

		COLON	IMEDD+5,AGAIN,AGAIN
                DC      COMPI,BRAN,COMMA,EXIT

;; -    IF          ( -- A )
;		Begin a conditional branch structure.

		COLON	IMEDD+2,IF,IFF
                DC      COMPI,QBRAN,HERE
                DC      DOLIT,0,COMMA,EXIT

;; -    AHEAD       ( -- A )
;		Compile a forward branch instruction.

		COLON	IMEDD+5,AHEAD,AHEAD
                DC      COMPI,BRAN,HERE,DOLIT,0,COMMA,EXIT

;; -    REPEAT      ( A a -- )
;		Terminate a BEGIN-WHILE-REPEAT indefinite loop.

		COLON	IMEDD+6,REPEAT,REPEA
                DC      AGAIN,HERE,SWAP,STORE,EXIT

;; -    THEN        ( A -- )
;		Terminate a conditional branch structure.

		COLON	IMEDD+4,THEN,THENN
                DC      HERE,SWAP,STORE,EXIT

;; -    AFT         ( a -- a A )
;		Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.

		COLON	IMEDD+3,AFT,AFT
                DC      DROP,AHEAD,BEGIN,SWAP,EXIT

;; -    ELSE        ( A -- A )
;		Start the false clause in an IF-ELSE-THEN structure.

		COLON	IMEDD+4,ELSE,ELSEE
                DC      AHEAD,SWAP,THENN,EXIT

;; -    WHILE       ( a -- A a )
;		Conditional branch out of a BEGIN-WHILE-REPEAT loop.

		COLON	IMEDD+5,WHILE,WHILE
                DC      IFF,SWAP,EXIT
;
;

;; -    ABORT"      ( -- ; <string> )
;               Conditional abort with an error message.

                HCODE   IMEDD+6,ABRTQ
                DC      'ABORT"'
                HCOLON
		DC	COMPI,ABORQ,STRCQ,EXIT


;; -    $"          ( -- ; <string> )
;		Compile an inline string literal.

                HCODE   IMEDD+2,STRQ
                DC      '$','"'
                HCOLON
		DC	COMPI,STRQP,STRCQ,EXIT

;; -    ."          ( -- ; <string> )
;		Compile an inline string literal to be typed out at run time.

                HCODE   IMEDD+2,DOTQ
                DC      '.','"'
                HCOLON
		DC	COMPI,DOTQP,STRCQ,EXIT



;;================================  Name compiler ================================

;; -    ?UNIQUE     ( a -- a )
;		Display a warning message if the word already exists.

		COLON	7,?UNIQUE,UNIQU
                DC      DUPP,NAMEQ              ;?name exists
                DC      QBRAN,UNIQ1             ;redefinitions are OK
		DC	DOTQP
		DC	6
		DC	' reDef'
                DC      OVER,COUNT,TYPEE        ;just in case its not planned
UNIQ1           DC      DROP,EXIT

;
;
;
;; -    $,n         ( na -- )
;		Build a new dictionary name using the string at na.

                HCODE   3,SNAME
                DC      '$',$2C,'n'
                HCOLON
		DC	DUPP,CAT		;?null input
		DC	QBRAN,PNAM1
		DC	UNIQU			;?redefinition
		DC	DUPP,LAST,STORE		;save na for vocabulary link
SNAM1           DC      HERE,ALGND,SWAP         ;align code address
                DC      CELLM                   ;link address
		DC	CRRNT,AT,AT,OVER,STORE
		DC	CELLM,DUPP,NP,STORE	;adjust name pointer
		DC	STORE,EXIT		;save code pointer
PNAM1		DC	STRQP
		DC	5
;		DC	' ','n','a','m','e'
		DC	' name'
		DC	0
		DC	0		;null input
		DC	THROW

;;================================  FORTH compiler ================================

;; -    $COMPILE    ( a -- )
;               Compile next word to code dictionary as a token or literal.
                COLON   8,$COMPILE,SCOMP
		DC	NAMEQ,QDUP		;?defined
		DC	QBRAN,SCOM2
		DC	AT,DOLIT,IMEDD,ANDD	;?immediate
		DC	QBRAN,SCOM1
		DC	EXECU,EXIT		;its immediate, execute
SCOM1		DC	COMMA,EXIT		;its not immediate, compile
SCOM2		DC	TNUMB,ATEXE		;try to convert to number
		DC	QBRAN,SCOM3
		DC	LITER,EXIT		;compile number as integer
SCOM3		DC	THROW			;error

;; -    OVERT       ( -- )
;		Link a new word into the current vocabulary.

		COLON	5,OVERT,OVERT
                DC      LAST,AT,CRRNT,AT,STORE,EXIT

;; -    ;           ( -- )
;		Terminate a colon definition.

		HCODE	IMEDD+COMPO+1,SEMIS
		DC	';'
		HCOLON
		DC	COMPI,EXIT,LBRAC,OVERT,EXIT

;; -    ]           ( -- )
;		Start compiling the words in the input stream.

		COLON	1,],RBRAC
                DC      DOLIT,SCOMP,TEVAL,STORE,EXIT

;; -    call,       ( ca -- )
;		Assemble a call instruction to ca.


		HCODE	5,CALLC
                DC      'call,'
		HCOLON
;                DC      DOLIT,CALLL,COMMA,HERE  ;Direct Threaded Code
;                DC      CELLP,SUBB,COMMA,EXIT

                DC      DOLIT,CALLL,DOLIT,DOLST
                DC      XORR,COMMA,EXIT

;; -    :           ( -- ; <string> )
;		Start a new colon definition using next word as its name.

		COLON	1,:,COLON
;                DC      TOKEN,SNAME,DOLIT,DOLST
;                DC      CALLC,RBRAC,EXIT

                DC      TOKEN,SNAME
                DC      CALLC,RBRAC,EXIT


;; -    IMMEDIATE   ( -- )
;		Make the last compiled word an immediate word.

		COLON	9,IMMEDIATE,IMMED
                DC      DOLIT,IMEDD,LAST,AT,AT,ORR
                DC      LAST,AT,STORE,EXIT

;;================================  Defining words ================================

;; -    USER        ( u -- ; <string> )
;		Compile a new user variable.

		COLON	4,USER,USER
                DC      TOKEN,SNAME,OVERT
                DC      DOLIT,DOLST,CALLC
                DC      COMPI,DOUSE,COMMA,EXIT

;; -    CREATE      ( -- ; <string> )
;		Compile a new array entry without allocating code space.

		COLON	6,CREATE,CREAT
                DC      TOKEN,SNAME,OVERT
                DC      DOLIT,DOLST,CALLC
                DC      COMPI,DOVAR,EXIT

;; -    VARIABLE    ( -- ; <string> )
;		Compile a new variable initialized to 0.

		COLON	8,VARIABLE,VARIA
                DC      CREAT,DOLIT,0,COMMA,EXIT

;;================================  Tools ================================

;; -    _TYPE       ( b u -- )
;               Display a string. Filter non-printing characters.

		COLON	5,_TYPE,UTYPE
                DC      TOR                     ;start count down loop
                DC      BRAN,UTYP2              ;skip first pass
UTYP1           DC      DUPP,CAT,TCHAR,EMIT     ;display only printable
                DC      DOLIT,1,PLUS            ;increment address
UTYP2           DC      DONXT,UTYP1             ;loop till done
                DC      DROP,EXIT



                COLON   7,PRNMEMT,PRNMEMT
                DC      MEMSPAC,AT
                DC      DUPP
                DC      DOLIT,'P',EQUAL
                DC      QBRAN,prnX
                DC      DROP,DOLIT,'P',EMIT
                DC      DOLIT,':',EMIT,EXIT
prnX            DC      DOLIT,'X',EQUAL
                DC      QBRAN,prnY
                DC      DOLIT,'X',EMIT,DOLIT,':',EMIT,EXIT
prnY            DC      DOLIT,'Y',EMIT,DOLIT,':',EMIT,EXIT



;;-     GETMDAT - get contents of x,y, or p memory
                COLON   7,GETMDAT,GETMDAT
                DC      MEMSPAC,AT
                DC      DUPP
                DC      DOLIT,'P',EQUAL
                DC      QBRAN,getX
                DC      DROP,PAT,EXIT
getX            DC      DOLIT,'X',EQUAL
                DC      QBRAN,getY
                DC      CAT,EXIT
getY            DC      YAT,EXIT

                COLON   5,FREGS,FREGS
                DC      CR,CR
                DC      DOTQP
                DC      9
                DC      'SP(r7) : '
                DC      SPAT,DOLIT,4,UDOTR
                DC      DOTQP
                DC      11
                DC      '  RP(r6) : '
                DC      RPAT,DOLIT,4,UDOTR
                DC      DOTQP
                DC      11
                DC      '  IP(r5) : '
                DC      IPAT,DOLIT,4,UDOTR
                DC      CR,CR,EXIT



;; -    dm+         ( a u -- a )
;               Dump u bytes from A, leaving a+u on the stack.

		COLON	3,dm+,DMP
                DC      PRNMEMT
                DC      OVER,DOLIT,4,UDOTR      ;display address
		DC	SPACE,TOR		;start count down loop
		DC	BRAN,PDUM2		;skip first pass
PDUM1           DC      DUPP
                DC      GETMDAT
                DC      DOLIT,7,UDOTR  ;display numeric data
;                DC      CAT,DOLIT,7,UDOTR  ;display numeric data
		DC	DOLIT,1,PLUS		;increment address
PDUM2		DC	DONXT,PDUM1		;loop till done
		DC	EXIT



                COLON   4,XMEM,XMEM
                DC      DOLIT,'X',MEMSPAC,STORE
                DC      EXIT

                COLON   4,YMEM,YMEM
                DC      DOLIT,'Y',MEMSPAC,STORE
                DC      EXIT

                COLON   4,PMEM,PMEM
                DC      DOLIT,'P',MEMSPAC,STORE
                DC      EXIT


;; -    DUMP        ( a u -- )
;		Dump u bytes from a, in a formatted manner.

                COLON   4,DUMP,DUMP
                DC      BASE,AT,TOR,HEX         ;save radix, set hex
                DC      DOLIT,8,SLASH          ;change count to lines
                DC      TOR                     ;start count down loop
DUMP1           DC      CR,DOLIT,8,DDUP
                DC      DMP             ;display numeric
                DC      ROT,ROT
                DC      SPACE,SPACE,UTYPE       ;display printable characters
                DC      NUFQ,INVER              ;user control
                DC      QBRAN,DUMP2
                DC      DONXT,DUMP1             ;loop till done
                DC      BRAN,DUMP3
DUMP2           DC      RFROM,DROP              ;cleanup loop stack, early exit
DUMP3           DC      DROP,RFROM,BASE,STORE   ;restore radix
                DC      EXIT



;; -    .S          ( ... -- ... )
;		Display the contents of the data stack.

		COLON	2,.S,DOTS
                DC      CR,DEPTH                ;stack depth
                DC      TOR                     ;start count down loop
                DC      BRAN,DOTS2              ;skip first pass
DOTS1           DC      RAT,PICK,DOT            ;index stack, display contents
DOTS2           DC      DONXT,DOTS1             ;loop till done
                DC      DOTQP
                DC      4
		DC	' <sp'
                DC      EXIT

;; -    !CSP        ( -- )
;		Save stack pointer in CSP for error checking.

		COLON	4,!CSP,STCSP
                DC      SPAT,CSP,STORE,EXIT     ;save pointer

;; -    ?CSP        ( -- )
;		Abort if stack pointer differs from that saved in CSP.

		COLON	4,?CSP,QCSP
                DC      SPAT,CSP,AT,XORR        ; compare pointers
		DC	ABORQ,6
;		DC	's','t','a','c','k','s'		;abort if different
		DC	'stacks'
                DC      EXIT

;; -    >NAME       ( ca -- na | F )
;		Convert code address to a name address.

		COLON	5,>NAME,TNAME
                DC      CRRNT                   ;vocabulary link
TNAM1           DC      CELLP,AT,QDUP           ;check all vocabularies
                DC      QBRAN,TNAM4
                DC      DDUP
TNAM2           DC      AT,DUPP                 ;?last word in a vocabulary
                DC      QBRAN,TNAM3
                DC      DDUP,NAMET,XORR         ;compare
                DC      QBRAN,TNAM3
                DC      CELLM                   ;continue with next word
                DC      BRAN,TNAM2
TNAM3           DC      SWAP,DROP,QDUP
                DC      QBRAN,TNAM1
                DC      SWAP,DROP,SWAP,DROP,EXIT
TNAM4           DC      DROP,DOLIT,0,EXIT       ;false flag

;; -    .ID         ( na -- )
;		Display the name at address.

		COLON	3,.ID,DOTID
                DC      QDUP                    ;if zero no name
                DC      QBRAN,DOTI1
                DC      COUNT,DOLIT,$01F,ANDD   ;mask lexicon bits
                DC      UTYPE,EXIT              ;display name string
DOTI1		DC	DOTQP
                DC      11
		DC	' {noName}  '
                DC      EXIT

;; -    SEE         ( -- ; <string> )
;		A simple decompiler.

		COLON	3,SEE,SEE
                DC      TICK                    ;starting address
                DC      CR,CELLP
SEE1            DC      CELLP,DUPP,AT,DUPP      ;?does it contain a zero
                DC      QBRAN,SEE2
                DC      TNAME                   ;?is it a name
SEE2            DC      QDUP                    ;name address or zero
                DC      QBRAN,SEE3
                DC      SPACE,DOTID             ;display name
                DC      BRAN,SEE4
SEE3            DC      DUPP,AT,UDOT            ;display number
SEE4            DC      NUFQ                    ;user control
                DC      QBRAN,SEE1
                DC      DROP,EXIT

;; -    WORDS       ( -- )
;		Display the names in the context vocabulary.

		COLON	5,WORDS,WORDS
                DC      CR,CNTXT,AT          ;only in context
WORS1           DC      AT,QDUP                 ;?at end of list
                DC      QBRAN,WORS2
                DC      DUPP,SPACE,DOTID        ;display a name
                DC      CELLM,NUFQ                              ;user control
                DC      QBRAN,WORS1
                DC      DROP
WORS2           DC      EXIT

;;================================  Hardware reset ================================

;; -    VER         ( -- n )


;		Return the version number of this implementation.

		COLON	3,VER,VERSN
                DC      DOLIT,VER*256+EXT,EXIT

;; -    hi          ( -- )
;		Display the sign-on message of eForth.

                COLON   2,hi,HI
                DC      DOLIT,SCIBAUD,STOIO,CR          ;initialize I/O
                DC      CR,DOTQP
                DC      15
                DC      'eForth 56k beta'
;                DC      VER+'0','.',EXT+'0'     ;version
                DC      CR,CR
                DC      EXIT



;; -    'BOOT       ( -- a )

;		The application startup vector.

		HCODE	5,TBOOT
                DC      $27
		DC	'BOOT'
		HCOLON
                DC      DOVAR
		DC	HI			;application to boot


;; -    COLD        ( -- )
;		The hilevel cold start sequence.

		COLON	4,COLD,COLD
COLD1           DC      DOLIT,UZERO,DOLIT,UPP
                DC      DOLIT,ULAST-UZERO,CMOVE
                DC      PRESE                   ;initialize stack and TIB
                DC      TBOOT,ATEXE             ;application boot
                DC      FORTH,CNTXT,AT,DUPP     ;initialize search order
                DC      CRRNT,DSTOR,OVERT
TAG2            DC      QUIT                    ;start interpretation
                DC      BRAN,COLD1              ;just in case

LASTW		nop

;===============================================================

LASTN           EQU     NAMEp+2                 ;last name address in name dictionary
NTOP            EQU     NAMEp-0                 ;next available memory in name dictionary

        if prnmsgs==1
        msg               "LASTN",'  ',LASTN
        msg               "NTOP",'  ',NTOP
        endif


CTOP		EQU	*+0			;next available memory in code dictionary



        ORG     P:RESET
        jmp     ORIG




;===============================================================